#include <torch/extension.h>
#include <vector>
#include <cmath>

// LSTM elementwise forward kernel with improved memory coalescing
__global__ void lstm_elementwise_forward(
    const float* __restrict__ gates,
    const float* __restrict__ prev_c,
    float* __restrict__ h,
    float* __restrict__ c,
    int batch_size,
    int hidden_size
) {
    int total = batch_size * hidden_size;
    // Grid-stride loop for full coverage
    for (int idx = blockIdx.x * blockDim.x + threadIdx.x; idx < total; idx += blockDim.x * gridDim.x) {
        int b = idx / hidden_size;
        int n = idx % hidden_size;
        // Precompute base pointer for the current batch element
        const float* gate_base = gates + b * hidden_size * 4;

        // Load gates from memory: note that the four gate values are stored with a fixed stride
        float i_gate = gate_base[n];
        float f_gate = gate_base[n + hidden_size];
        float g_gate = gate_base[n + 2 * hidden_size];
        float o_gate = gate_base[n + 3 * hidden_size];

        // Apply activations (sigmoid and tanh)
        i_gate = 1.f / (1.f + expf(-i_gate));
        f_gate = 1.f / (1.f + expf(-f_gate));
        g_gate = tanhf(g_gate);
        o_gate = 1.f / (1.f + expf(-o_gate));

        // Update cell state and hidden state
        float c_new = f_gate * prev_c[idx] + i_gate * g_gate;
        float h_new = o_gate * tanhf(c_new);

        c[idx] = c_new;
        h[idx] = h_new;
    }
}

// Linear forward kernel with grid-stride loop for coalesced global memory accesses
__global__ void linear_forward_kernel(
    const float* __restrict__ input,
    const float* __restrict__ weight,
    const float* __restrict__ bias,
    float* __restrict__ output,
    int batch_size,
    int input_dim,
    int output_dim
) {
    int total = batch_size * output_dim;
    for (int idx = blockIdx.x * blockDim.x + threadIdx.x; idx < total; idx += blockDim.x * gridDim.x) {
        int batch_idx = idx / output_dim;
        int out_idx = idx % output_dim;
        float sum = 0.f;
        int base_input = batch_idx * input_dim;
        int base_weight = out_idx * input_dim;
        for (int k = 0; k < input_dim; k++) {
            sum += input[base_input + k] * weight[base_weight + k];
        }
        if (bias != nullptr) {
            sum += bias[out_idx];
        }
        output[idx] = sum;
    }
}

// Function to compute the forward pass of an LSTM layer using the lstm_elementwise_forward kernel
torch::Tensor lstm_forward_cuda(
    torch::Tensor input,
    torch::Tensor w_ih,
    torch::Tensor w_hh,
    torch::Tensor b_ih,
    torch::Tensor b_hh,
    torch::Tensor h0,
    torch::Tensor c0
) {
    int batch_size = input.size(0);
    int seq_len = input.size(1);
    int input_size = input.size(2);
    int hidden_size = h0.size(1);

    torch::Tensor h = h0.clone();
    torch::Tensor c = c0.clone();
    std::vector<torch::Tensor> outputs;

    for (int t = 0; t < seq_len; t++) {
        // Get input at timestep t
        torch::Tensor xt = input.select(1, t);

        // Compute gates: gates = xt * w_ih^T + h * w_hh^T + b_ih + b_hh
        torch::Tensor gates = torch::addmm(b_ih, xt, w_ih.t());
        gates = torch::addmm(gates, h, w_hh.t());
        gates += b_hh;

        const float* gates_ptr = gates.data_ptr<float>();
        const float* prev_c_ptr = c.data_ptr<float>();
        float* h_ptr = h.data_ptr<float>();
        float* c_ptr = c.data_ptr<float>();

        int total_elements = batch_size * hidden_size;
        int threads = 256;
        int blocks = (total_elements + threads - 1) / threads;

        lstm_elementwise_forward<<<blocks, threads>>>(
            gates_ptr,
            prev_c_ptr,
            h_ptr,
            c_ptr,
            batch_size,
            hidden_size
        );

        outputs.push_back(h.unsqueeze(1));
    }

    return torch::cat(outputs, 1);
}

// Linear forward pass function that invokes the linear_forward_kernel
torch::Tensor linear_forward_cuda(
    torch::Tensor input,
    torch::Tensor weight,
    torch::Tensor bias
) {
    int batch_size = input.size(0);
    int input_dim = input.size(1);
    int output_dim = weight.size(0);
    auto options = torch::TensorOptions().dtype(input.dtype()).device(input.device());
    torch::Tensor output = torch::empty({batch_size, output_dim}, options);

    int threads = 256;
    int total = batch_size * output_dim;
    int blocks = (total + threads - 1) / threads;
    
    linear_forward_kernel<<<blocks, threads>>>(
         input.data_ptr<float>(),
         weight.data_ptr<float>(),
         bias.defined() ? bias.data_ptr<float>() : nullptr,
         output.data_ptr<float>(),
         batch_size,
         input_dim,
         output_dim
    );
    
    return output;
}

// Main forward function orchestrating multi-layer LSTM and final linear layer
torch::Tensor forward(
    torch::Tensor x,
    std::vector<torch::Tensor> lstm_weights_ih,
    std::vector<torch::Tensor> lstm_weights_hh,
    std::vector<torch::Tensor> lstm_biases_ih,
    std::vector<torch::Tensor> lstm_biases_hh,
    torch::Tensor fc_weight,
    torch::Tensor fc_bias,
    torch::Tensor h0,
    torch::Tensor c0,
    bool is_training
) {
    h0 = h0.to(x.device());
    c0 = c0.to(x.device());
    torch::Tensor out = x;
    int num_layers = lstm_weights_ih.size();

    for (int i = 0; i < num_layers; i++) {
         auto w_ih = lstm_weights_ih[i].to(x.device());
         auto w_hh = lstm_weights_hh[i].to(x.device());
         auto b_ih = lstm_biases_ih[i].to(x.device());
         auto b_hh = lstm_biases_hh[i].to(x.device());
         
         // Select corresponding initial hidden and cell states
         torch::Tensor h_i = h0.narrow(0, i, 1).squeeze(0);
         torch::Tensor c_i = c0.narrow(0, i, 1).squeeze(0);
         
         out = lstm_forward_cuda(out, w_ih, w_hh, b_ih, b_hh, h_i, c_i);
    }

    // Select the last timestep's output and apply the final linear layer
    out = out.select(1, -1);
    out = linear_forward_cuda(out, fc_weight, fc_bias);
    
    return out;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Module forward (CUDA) with aligned memory accesses for coalescing");
}
